/*

Copyright (c) 2022 D.S. Andreeff

This file is part of WinClock.

WinClock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

WinClock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with WinClock. If not, see <http://www.gnu.org/licenses/>.

*/

#include <windows.h>
#include <windowsx.h>

#include <stdio.h>

#define SEG_1 ( 1 << 0 )
#define SEG_2 ( 1 << 1 )
#define SEG_3 ( 1 << 2 )
#define SEG_4 ( 1 << 3 )
#define SEG_5 ( 1 << 4 )
#define SEG_6 ( 1 << 5 )
#define SEG_7 ( 1 << 6 )
#define SEG_8 ( 1 << 7 )

typedef struct {
	char key;
	unsigned char value;
} digit_segs_entry;

digit_segs_entry digit_segs_map[] = {
	{ '0', ( SEG_1 | SEG_2 | SEG_3 | SEG_4 | SEG_5 | SEG_6 ) },
	{ '1', ( SEG_2 | SEG_3 ) },
	{ '2', ( SEG_1 | SEG_6 | SEG_7 | SEG_3 | SEG_4 ) },
	{ '3', ( SEG_1 | SEG_2 | SEG_7 | SEG_3 | SEG_4 ) },
	{ '4', ( SEG_2 | SEG_5 | SEG_7 | SEG_3 ) },
	{ '5', ( SEG_1 | SEG_2 | SEG_7 | SEG_5 | SEG_4 ) },
	{ '6', ( SEG_4 | SEG_5 | SEG_6 | SEG_1 | SEG_2 | SEG_7 ) },
	{ '7', ( SEG_4 | SEG_3 | SEG_2 ) },
	{ '8', ( SEG_1 | SEG_2 | SEG_3 | SEG_4 | SEG_5 | SEG_6 | SEG_7 ) },
	{ '9', ( SEG_1 | SEG_2 | SEG_3 | SEG_4 | SEG_5 | SEG_7 ) },
	{ ':', ( SEG_8 ) },
	};

unsigned char segs_from_digit(char digit) {
	const int map_length = sizeof(digit_segs_map) / sizeof(digit_segs_map[0]);

	int current; for(current=0; current<map_length; current++) {
		digit_segs_entry entry = digit_segs_map[current];
		if(entry.key == digit) { return entry.value; }
		}

	return 0;
	}

typedef struct {
	unsigned char key;
	float x_start, y_start;
	float x_finish, y_finish;
} segs_offsets_entry;

segs_offsets_entry segs_offset_map[] = {
	{ SEG_1, 0.0, 2.0, 1.0, 2.0 },
	{ SEG_2, 1.0, 2.0, 1.0, 1.0 },
	{ SEG_3, 1.0, 1.0, 1.0, 0.0 },
	{ SEG_4, 0.0, 0.0, 1.0, 0.0 },
	{ SEG_5, 0.0, 1.0, 0.0, 0.0 },
	{ SEG_6, 0.0, 2.0, 0.0, 1.0 },
	{ SEG_7, 0.0, 1.0, 1.0, 1.0 },
	{ SEG_8, 0.5, 1.5, 0.5, 0.5 },
	};

const int segment_size = 32;
const int segment_margin = 16;
const int segment_width = 8;

BOOL Debug(LPSTR debug_line,HWND wnd) {
	const char class_name[] = "DebugBoardClass";
	const char window_name[] = "DebugBoardWindow";

	HWND server_window = FindWindow(class_name,window_name);
	if(server_window == NULL) { return FALSE; }

	COPYDATASTRUCT data;
	ZeroMemory(&data,sizeof(data));
	data.lpData = debug_line;
	data.cbData = strlen(debug_line)+1;

	LRESULT send_result = SendMessage(server_window,WM_COPYDATA,(WPARAM)wnd,(LPARAM)&data);
	if(send_result != TRUE) { return FALSE; }

	return TRUE;
	}

char date_buffer[256];
char time_buffer[256];

void UpdateDateTime(HWND wnd) {
	const char * weekdays[] = { "", "", "", "", "", "", "" };
	const int weekdays_l = sizeof(weekdays) / sizeof(weekdays[0]);

	SYSTEMTIME dt;
	GetLocalTime(&dt);

	ZeroMemory(date_buffer,sizeof(date_buffer));
	ZeroMemory(time_buffer,sizeof(time_buffer));

	const char * weekday = weekdays[ dt.wDayOfWeek % weekdays_l ];

	sprintf(date_buffer,"%02d.%02d.%04d (%s)",dt.wDay,dt.wMonth,dt.wYear,weekday);
	sprintf(time_buffer,"%02d:%02d:%02d",dt.wHour,dt.wMinute,dt.wSecond);

	SetWindowText(wnd,date_buffer);

	InvalidateRect(wnd,NULL,FALSE);
	UpdateWindow(wnd);
	}

void RepaintWindow(HDC dc,RECT rect) {
	FillRect(dc,&rect,(HBRUSH)GetStockObject(WHITE_BRUSH));

	HPEN pen_lit = CreatePen(PS_SOLID,segment_width,RGB(0,0,0));
	HPEN pen_dim = CreatePen(PS_SOLID,segment_width,RGB(224,224,224));

	const int time_buffer_l = strlen(time_buffer);
	for(int char_num=0; char_num<time_buffer_l; char_num++) {
		char digit = time_buffer[ char_num % time_buffer_l ];
		unsigned char segs = segs_from_digit(digit);

		int base = segment_margin + ( segment_size + segment_margin ) * char_num;
		const int map_length = sizeof(segs_offset_map) / sizeof(segs_offset_map[0]);

		int seg_num;

		SelectObject(dc,pen_dim);
		for(seg_num=0; seg_num<map_length; seg_num++) {
			segs_offsets_entry entry = segs_offset_map[seg_num];
			if((segs & entry.key) ? 1 : 0) { continue; }

			int start_x = base + segment_size * entry.x_start;
			int start_y = segment_margin + segment_size * entry.y_start;
			MoveToEx(dc,start_x,start_y,NULL);

			int finish_x = base + segment_size * entry.x_finish;
			int finish_y = segment_margin + segment_size * entry.y_finish;
			LineTo(dc,finish_x,finish_y);
			}

		SelectObject(dc,pen_lit);
		for(seg_num=0; seg_num<map_length; seg_num++) {
			segs_offsets_entry entry = segs_offset_map[seg_num];
			if((segs & entry.key) ? 0 : 1) { continue; }

			int start_x = base + segment_size * entry.x_start;
			int start_y = segment_margin + segment_size * entry.y_start;
			MoveToEx(dc,start_x,start_y,NULL);

			int finish_x = base + segment_size * entry.x_finish;
			int finish_y = segment_margin + segment_size * entry.y_finish;
			LineTo(dc,finish_x,finish_y);
			}
/*
		for(int seg_num=0; seg_num<map_length; seg_num++) {
			segs_offsets_entry entry = segs_offset_map[seg_num];
			SelectObject(dc,(segs & entry.key) ? pen_lit : pen_dim);

			int start_x = base + segment_size * entry.x_start;
			int start_y = segment_margin + segment_size * entry.y_start;
			MoveToEx(dc,start_x,start_y,NULL);

			int finish_x = base + segment_size * entry.x_finish;
			int finish_y = segment_margin + segment_size * entry.y_finish;
			LineTo(dc,finish_x,finish_y);
			}
*/
		}

	DeleteObject(pen_lit);
	DeleteObject(pen_dim);
	}

LRESULT CALLBACK WindowProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp) { switch(msg) {
	case WM_CREATE:
		UpdateDateTime(wnd);
		return 0;

	case WM_TIMER:
		UpdateDateTime(wnd);
		return 0;

	case WM_PAINT: {
		RECT rect;
		GetClientRect(wnd,&rect);

		PAINTSTRUCT window_ps;
		HDC window_dc = BeginPaint(wnd,&window_ps);

//		RepaintWindow(window_dc,rect);

		HDC buffer_dc = CreateCompatibleDC(window_dc);
		if(buffer_dc != NULL) {
			HBITMAP buffer_bitmap = CreateCompatibleBitmap(window_dc,rect.right,rect.bottom);
			if(buffer_bitmap != NULL) {
				HGDIOBJ old_object = SelectObject(buffer_dc,buffer_bitmap);

				RepaintWindow(buffer_dc,rect);
				BitBlt(window_dc,0,0,rect.right,rect.bottom,buffer_dc,0,0,SRCCOPY);

				SelectObject(old_object,buffer_bitmap);

				DeleteObject(buffer_bitmap);
				}

			DeleteDC(buffer_dc);
			}

		EndPaint(wnd,&window_ps);
		} return 0;

	case WM_CLOSE:
		DestroyWindow(wnd);
		return 0;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;

	default:
		return DefWindowProc(wnd,msg,wp,lp);
	} }

int PASCAL WinMain(HINSTANCE instance,HINSTANCE,LPSTR,int) {
	const char class_name[] = "ClockClass";
	const char window_name[] = "ClockWindow";

	WNDCLASS window_class;
	window_class.style = CS_HREDRAW | CS_VREDRAW;
	window_class.lpfnWndProc = WindowProc;
	window_class.cbClsExtra = 0;
	window_class.cbWndExtra = 0;
	window_class.hInstance = instance;
	window_class.hIcon = LoadIcon(instance,MAKEINTRESOURCE(1));
	window_class.hCursor = LoadCursor(NULL,IDC_ARROW);
	window_class.hbrBackground = NULL;
	window_class.lpszMenuName = NULL;
	window_class.lpszClassName = class_name;

	if(RegisterClass(&window_class) == 0) { return -1; }

	const int window_width = segment_margin + ( segment_size + segment_margin ) * 8;
	const int window_height = ( segment_margin + segment_size ) * 2;

	const int screen_width = GetSystemMetrics(SM_CXSCREEN);
	const int screen_height = GetSystemMetrics(SM_CYSCREEN);

	DWORD window_style = WS_DLGFRAME|WS_BORDER|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_VISIBLE;

	RECT window_rect;
	window_rect.left = ( screen_width - window_width ) / 2;
	window_rect.top = ( screen_height - window_height ) / 2;
	window_rect.right = window_rect.left + window_width;
	window_rect.bottom = window_rect.top + window_height;

	if(AdjustWindowRect(&window_rect,window_style,FALSE) == FALSE) { return -1; }

	HWND main_window = CreateWindowEx(WS_EX_TOPMOST,class_name,window_name,window_style,
		window_rect.left,window_rect.top,window_rect.right-window_rect.left,window_rect.bottom-window_rect.top,
		NULL,NULL,instance,NULL);
	if(main_window == 0) { return -1; }

	if(SetTimer(main_window,-1,1000,NULL) == 0) { return -1; }

	MSG msg;
	while(GetMessage(&msg,NULL,0,0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		}

	if(KillTimer(main_window,-1) == FALSE) { return -1; }

	return msg.wParam;
	}

